import asyncio
import logging
import os
import shutil
import struct
from pathlib import Path

logger = logging.getLogger(__name__)

BLIT_SOCKET_ENV = "INDENT_TERMINAL_CONTROLLER_BLIT_SOCKET"
BLIT_SERVER_BINARY = "blit-server"
_SOCKET_READY_RETRIES = 50
_SOCKET_READY_INTERVAL_S = 0.1
_BLIT_FRAME_TIMEOUT_S = 0.25
_BLIT_CREATE_TIMEOUT_S = 1.5
_TAGGED_CREATE_MAX_RETRIES = 5
_TAGGED_CREATE_INITIAL_BACKOFF_S = 0.2

C2S_CREATE = 0x10
S2C_CREATED = 0x01
S2C_LIST = 0x03


def _encode_frame(payload: bytes) -> bytes:
    return struct.pack("<I", len(payload)) + payload


def _build_tagged_create_message(
    rows: int,
    cols: int,
    tag: str,
    command: str,
) -> bytes:
    tag_bytes = tag.encode("utf-8")
    command_bytes = command.encode("utf-8")
    return (
        bytes([C2S_CREATE])
        + rows.to_bytes(2, "little")
        + cols.to_bytes(2, "little")
        + len(tag_bytes).to_bytes(2, "little")
        + tag_bytes
        + command_bytes
    )


async def _read_frame(
    reader: asyncio.StreamReader,
    timeout_s: float,
) -> bytes | None:
    try:
        length_buf = await asyncio.wait_for(reader.readexactly(4), timeout=timeout_s)
        length = struct.unpack("<I", length_buf)[0]
        if length == 0:
            return b""
        return await asyncio.wait_for(reader.readexactly(length), timeout=timeout_s)
    except (asyncio.IncompleteReadError, ConnectionError, OSError, TimeoutError):
        return None


async def _read_until_initial_list(
    reader: asyncio.StreamReader,
    timeout_s: float,
) -> bytes | None:
    while True:
        frame = await _read_frame(reader, timeout_s=timeout_s)
        if frame is None:
            return None
        if frame and frame[0] == S2C_LIST:
            return frame


def get_blit_socket_path(terminal_controller_socket_path: str) -> str:
    explicit_socket_path = os.environ.get(BLIT_SOCKET_ENV)
    if explicit_socket_path:
        return explicit_socket_path

    socket_path = Path(terminal_controller_socket_path)
    if socket_path.suffix:
        return str(socket_path.with_name(f"{socket_path.stem}-blit{socket_path.suffix}"))
    return f"{terminal_controller_socket_path}-blit"


class BlitTerminalServer:
    def __init__(self, socket_path: str):
        self.socket_path = socket_path
        self._process: asyncio.subprocess.Process | None = None
        self._stdout_task: asyncio.Task[None] | None = None
        self._stderr_task: asyncio.Task[None] | None = None

    @classmethod
    def from_terminal_controller_socket(cls, terminal_controller_socket_path: str) -> "BlitTerminalServer":
        return cls(socket_path=get_blit_socket_path(terminal_controller_socket_path))

    async def start(self) -> None:
        if self._process is not None:
            raise RuntimeError("Blit terminal server already started")

        blit_server_binary = shutil.which(BLIT_SERVER_BINARY)
        if blit_server_binary is None:
            raise RuntimeError(f"{BLIT_SERVER_BINARY} was not found on PATH")

        env = os.environ.copy()
        env["BLIT_SOCK"] = self.socket_path
        env.setdefault("SHELL", os.environ.get("SHELL", "/bin/sh"))

        self._process = await asyncio.create_subprocess_exec(
            blit_server_binary,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            env=env,
        )

        if self._process.stdout is not None:
            self._stdout_task = asyncio.create_task(
                self._log_stream(self._process.stdout, logging.INFO, "blit-server stdout")
            )
        if self._process.stderr is not None:
            self._stderr_task = asyncio.create_task(
                self._log_stream(self._process.stderr, logging.WARNING, "blit-server stderr")
            )

        await self._wait_for_socket()
        logger.info("Blit terminal server ready", extra={"socket_path": self.socket_path})

    async def stop(self) -> None:
        tasks = [task for task in (self._stdout_task, self._stderr_task) if task is not None]
        process = self._process

        if process is not None and process.returncode is None:
            process.terminate()
            try:
                await asyncio.wait_for(process.wait(), timeout=5)
            except TimeoutError:
                process.kill()
                await process.wait()

        for task in tasks:
            task.cancel()
        if tasks:
            await asyncio.gather(*tasks, return_exceptions=True)

        self._stdout_task = None
        self._stderr_task = None
        self._process = None

    async def start_tagged_terminal(
        self,
        *,
        rows: int,
        cols: int,
        tag: str,
        command: str,
    ) -> int:
        last_error: Exception | None = None
        for attempt in range(_TAGGED_CREATE_MAX_RETRIES):
            try:
                return await self._try_start_tagged_terminal(
                    rows=rows,
                    cols=cols,
                    tag=tag,
                    command=command,
                )
            except Exception as exc:
                last_error = exc
                if attempt < _TAGGED_CREATE_MAX_RETRIES - 1:
                    backoff = _TAGGED_CREATE_INITIAL_BACKOFF_S * (2**attempt)
                    logger.warning(
                        "Blit tagged terminal creation failed, retrying",
                        extra={
                            "attempt": attempt + 1,
                            "max_retries": _TAGGED_CREATE_MAX_RETRIES,
                            "backoff_s": backoff,
                            "error": str(exc),
                            "socket_path": self.socket_path,
                        },
                    )
                    await asyncio.sleep(backoff)
        raise RuntimeError(
            f"Failed to create blit terminal after {_TAGGED_CREATE_MAX_RETRIES} attempts on {self.socket_path}"
        ) from last_error

    async def _try_start_tagged_terminal(
        self,
        *,
        rows: int,
        cols: int,
        tag: str,
        command: str,
    ) -> int:
        reader, writer = await asyncio.open_unix_connection(self.socket_path)
        try:
            list_frame = await _read_until_initial_list(reader, timeout_s=_BLIT_FRAME_TIMEOUT_S)
            if list_frame is None:
                raise RuntimeError(f"Timed out waiting for initial blit list on {self.socket_path}")

            writer.write(
                _encode_frame(
                    _build_tagged_create_message(
                        rows=rows,
                        cols=cols,
                        tag=tag,
                        command=command,
                    )
                )
            )
            await writer.drain()

            while True:
                frame = await _read_frame(reader, timeout_s=_BLIT_CREATE_TIMEOUT_S)
                if frame is None:
                    raise RuntimeError(f"Timed out waiting for blit terminal creation on {self.socket_path}")
                if len(frame) >= 3 and frame[0] == S2C_CREATED:
                    return int.from_bytes(frame[1:3], "little")
        finally:
            writer.close()
            await writer.wait_closed()

    async def _wait_for_socket(self) -> None:
        for _ in range(_SOCKET_READY_RETRIES):
            if os.path.exists(self.socket_path):
                return

            process = self._process
            if process is not None and process.returncode is not None:
                raise RuntimeError(f"blit terminal server exited before creating socket {self.socket_path}")

            await asyncio.sleep(_SOCKET_READY_INTERVAL_S)

        raise RuntimeError(f"Timed out waiting for blit terminal socket {self.socket_path}")

    async def _log_stream(
        self,
        stream: asyncio.StreamReader,
        level: int,
        label: str,
    ) -> None:
        try:
            while True:
                raw_line = await stream.readline()
                if not raw_line:
                    return
                logger.log(level, "%s: %s", label, raw_line.decode("utf-8", errors="replace").rstrip())
        except asyncio.CancelledError:
            raise
        except Exception:
            logger.debug("Failed to read blit terminal server logs", exc_info=True)



from __future__ import annotations

import re
from collections.abc import Callable
from dataclasses import dataclass
from pathlib import Path, PurePosixPath

from anyio import Path as AsyncPath

from exponent.core.remote_execution.cli_rpc_types import ApplyPatchToolInput, ApplyPatchToolResult
from exponent.core.remote_execution.utils import safe_read_file

_HUNK_HEADER_RE = re.compile(
    r"^@@ -(?P<old_start>\d+)(?:,(?P<old_count>\d+))? \+(?P<new_start>\d+)(?:,(?P<new_count>\d+))? @@(?: .*)?$"
)
_DIFF_HEADER_PREFIXES = ("diff --git ", "index ", "--- ", "+++ ")
_CONTEXT_HUNK_PREFIX = "@@ "
_EMPTY_CONTEXT_HUNK = "@@"
_END_OF_FILE_MARKER = "*** End of File"
_NO_NEWLINE_MARKER = r"\ No newline at end of file"


class ApplyPatchError(ValueError):
    pass


@dataclass
class _DiffLine:
    prefix: str
    text: str
    has_newline: bool = True


@dataclass(frozen=True)
class _UnifiedDiffHunk:
    old_start: int
    old_count: int
    new_start: int
    new_count: int
    lines: tuple[_DiffLine, ...]


@dataclass(frozen=True)
class _ContextDiffHunk:
    change_context: str | None
    lines: tuple[_DiffLine, ...]
    is_end_of_file: bool = False


@dataclass(frozen=True)
class _FileLine:
    text: str
    has_newline: bool


async def execute_apply_patch_tool(
    input_data: ApplyPatchToolInput,
    working_directory: str,
) -> ApplyPatchToolResult:
    try:
        display_path, absolute_path = _resolve_patch_paths(input_data.path, working_directory)
    except ApplyPatchError as exc:
        return ApplyPatchToolResult(status="failed", output=str(exc))

    if input_data.operation_type == "create_file":
        return await _create_file(display_path, absolute_path, input_data.diff or "")
    if input_data.operation_type == "update_file":
        return await _update_file(display_path, absolute_path, input_data.diff or "")
    if input_data.operation_type == "delete_file":
        return await _delete_file(display_path, absolute_path)
    return ApplyPatchToolResult(
        status="failed",
        output=f"apply_patch received an unknown operation type: {input_data.operation_type}",
    )


async def _create_file(
    display_path: str,
    absolute_path: Path,
    diff: str,
) -> ApplyPatchToolResult:
    try:
        exists = await AsyncPath(absolute_path).exists()
    except (OSError, PermissionError) as exc:
        return ApplyPatchToolResult(
            status="failed",
            output=f"Could not inspect '{display_path}' before applying the patch: {exc!s}",
        )
    if exists:
        return ApplyPatchToolResult(
            status="failed",
            output=f"Cannot create '{display_path}' because it already exists.",
        )

    try:
        new_content = apply_unified_diff(original_text="", diff=diff)
    except ApplyPatchError as exc:
        return ApplyPatchToolResult(
            status="failed",
            output=f"Could not build '{display_path}' from the patch: {exc!s}",
        )

    try:
        await _ensure_parent_dir(absolute_path)
        await AsyncPath(absolute_path).write_text(new_content)
    except OSError as exc:
        return ApplyPatchToolResult(status="failed", output=f"Could not write '{display_path}': {exc!s}")
    return ApplyPatchToolResult(status="completed", output=f"Applied patch to create '{display_path}'.")


async def _update_file(
    display_path: str,
    absolute_path: Path,
    diff: str,
) -> ApplyPatchToolResult:
    original_text_or_error = await _read_existing_file(display_path, absolute_path)
    if isinstance(original_text_or_error, ApplyPatchToolResult):
        return original_text_or_error

    try:
        new_content = apply_unified_diff(original_text=original_text_or_error, diff=diff)
    except ApplyPatchError as exc:
        return ApplyPatchToolResult(
            status="failed",
            output=f"Could not apply the patch to '{display_path}': {exc!s}",
        )

    try:
        await AsyncPath(absolute_path).write_text(new_content)
    except OSError as exc:
        return ApplyPatchToolResult(status="failed", output=f"Could not write '{display_path}': {exc!s}")
    return ApplyPatchToolResult(status="completed", output=f"Applied patch to update '{display_path}'.")


async def _delete_file(
    display_path: str,
    absolute_path: Path,
) -> ApplyPatchToolResult:
    path = AsyncPath(absolute_path)
    try:
        exists = await path.exists()
    except (OSError, PermissionError) as exc:
        return ApplyPatchToolResult(
            status="failed",
            output=f"Could not inspect '{display_path}' before applying the patch: {exc!s}",
        )
    if not exists:
        return ApplyPatchToolResult(
            status="failed",
            output=f"Cannot delete '{display_path}' because it does not exist.",
        )

    try:
        await path.unlink()
    except OSError as exc:
        return ApplyPatchToolResult(status="failed", output=f"Could not delete '{display_path}': {exc!s}")
    return ApplyPatchToolResult(status="completed", output=f"Applied patch to delete '{display_path}'.")


async def _read_existing_file(display_path: str, absolute_path: Path) -> str | ApplyPatchToolResult:
    path = AsyncPath(absolute_path)
    try:
        exists = await path.exists()
    except (OSError, PermissionError) as exc:
        return ApplyPatchToolResult(
            status="failed",
            output=f"Could not read '{display_path}' before applying the patch: {exc!s}",
        )

    if not exists:
        return ApplyPatchToolResult(
            status="failed",
            output=f"Could not read '{display_path}' before applying the patch: File not found",
        )

    try:
        if await path.is_dir():
            return ApplyPatchToolResult(
                status="failed",
                output=f"Could not read '{display_path}' before applying the patch: {absolute_path} is a directory",
            )
    except (OSError, PermissionError) as exc:
        return ApplyPatchToolResult(
            status="failed",
            output=f"Could not read '{display_path}' before applying the patch: {exc!s}",
        )

    try:
        return await safe_read_file(path)
    except PermissionError:
        return ApplyPatchToolResult(
            status="failed",
            output=f"Could not read '{display_path}' before applying the patch: Permission denied: cannot read {absolute_path}",
        )
    except UnicodeDecodeError:
        return ApplyPatchToolResult(
            status="failed",
            output=f"Could not read '{display_path}' before applying the patch: File appears to be binary or has invalid text encoding",
        )
    except Exception as exc:
        return ApplyPatchToolResult(
            status="failed",
            output=f"Could not read '{display_path}' before applying the patch: Error reading file: {exc!s}",
        )


async def _ensure_parent_dir(path: Path) -> None:
    parent = path.parent
    if str(parent) != ".":
        await AsyncPath(parent).mkdir(parents=True, exist_ok=True)


def _resolve_patch_paths(raw_path: str, working_directory: str) -> tuple[str, Path]:
    candidate = raw_path.strip()
    if not candidate:
        raise ApplyPatchError("apply_patch paths must not be empty.")
    if "\\" in candidate:
        raise ApplyPatchError("apply_patch paths must use forward slashes.")

    normalized_path = PurePosixPath(candidate)
    if normalized_path.is_absolute():
        display_path = normalized_path.as_posix()
        return display_path, Path(display_path)

    display_path = _validate_workspace_relative_path(candidate)
    return display_path, Path(working_directory, display_path)


def _validate_workspace_relative_path(path: str) -> str:
    candidate = path.strip()
    if not candidate:
        raise ApplyPatchError("apply_patch paths must not be empty.")
    if "\\" in candidate:
        raise ApplyPatchError("apply_patch paths must use forward slashes.")

    normalized_path = PurePosixPath(candidate)
    if normalized_path.is_absolute():
        raise ApplyPatchError("apply_patch paths must be relative to the workspace root.")

    cleaned_parts = []
    for part in normalized_path.parts:
        if part in {"", "."}:
            continue
        if part == "..":
            raise ApplyPatchError("apply_patch paths must stay within the workspace root.")
        cleaned_parts.append(part)

    if not cleaned_parts:
        raise ApplyPatchError("apply_patch paths must point to a file inside the workspace root.")

    return PurePosixPath(*cleaned_parts).as_posix()


def apply_unified_diff(*, original_text: str, diff: str) -> str:
    if not diff.strip():
        return original_text

    lines = _normalize_diff_lines(diff)
    diff_format = _detect_diff_format(lines)
    if diff_format == "unified":
        hunks = _parse_unified_diff_lines(lines)
        if not hunks:
            return original_text
        return _apply_unified_hunks(original_text=original_text, hunks=hunks)
    if diff_format == "context":
        hunks = _parse_context_diff_lines(lines)
        if not hunks:
            return original_text
        return _apply_context_hunks(original_text=original_text, hunks=hunks)
    return original_text


def _normalize_diff_lines(diff: str) -> list[str]:
    return [line.rstrip("\r") for line in diff.splitlines()]


def _detect_diff_format(lines: list[str]) -> str | None:
    for line in lines:
        if not line or line.startswith(_DIFF_HEADER_PREFIXES):
            continue
        if _HUNK_HEADER_RE.match(line):
            return "unified"
        if _is_context_hunk_header(line):
            return "context"
        raise ApplyPatchError(f"Unsupported apply_patch diff line: {line}")
    return None


def _apply_unified_hunks(*, original_text: str, hunks: list[_UnifiedDiffHunk]) -> str:
    original_lines = _split_file_lines(original_text)
    output_lines: list[_FileLine] = []
    source_index = 0

    for hunk in hunks:
        expected_source_index = _resolve_hunk_source_index(hunk)
        if expected_source_index < source_index:
            raise ApplyPatchError("apply_patch hunks overlap or are out of order.")
        if expected_source_index > len(original_lines):
            raise ApplyPatchError("apply_patch hunk starts past the end of the file.")

        output_lines.extend(original_lines[source_index:expected_source_index])
        source_index = expected_source_index
        hunk_source_index = source_index

        for diff_line in hunk.lines:
            if diff_line.prefix == " ":
                actual_line = _get_source_line(original_lines, source_index)
                _assert_line_matches(actual_line, diff_line.text)
                output_lines.append(actual_line)
                source_index += 1
            elif diff_line.prefix == "-":
                actual_line = _get_source_line(original_lines, source_index)
                _assert_line_matches(actual_line, diff_line.text)
                source_index += 1
            elif diff_line.prefix == "+":
                output_lines.append(_FileLine(text=diff_line.text, has_newline=diff_line.has_newline))
            else:
                raise ApplyPatchError(f"Unsupported unified diff line prefix: {diff_line.prefix}")

        consumed_old = source_index - hunk_source_index
        produced_new = sum(1 for line in hunk.lines if line.prefix != "-")
        if consumed_old != hunk.old_count:
            raise ApplyPatchError("apply_patch old-line count did not match the diff header.")
        if produced_new != hunk.new_count:
            raise ApplyPatchError("apply_patch new-line count did not match the diff header.")

    output_lines.extend(original_lines[source_index:])
    return _render_file_lines(output_lines)


def _apply_context_hunks(*, original_text: str, hunks: list[_ContextDiffHunk]) -> str:
    output_lines = _split_file_lines(original_text)
    search_index = 0

    for hunk in hunks:
        if hunk.change_context is not None:
            context_index = _seek_sequence(output_lines, [hunk.change_context], search_index, False)
            if context_index is None:
                raise ApplyPatchError(f"apply_patch could not find context line: {hunk.change_context!r}")
            search_index = context_index + 1

        pattern = [line.text for line in hunk.lines if line.prefix != "+"]
        match_index = _seek_sequence(output_lines, pattern, search_index, hunk.is_end_of_file)
        if match_index is None:
            if not pattern:
                raise ApplyPatchError("apply_patch context hunk could not determine where to insert new lines.")
            raise ApplyPatchError(
                "apply_patch diff did not match the current file contents.\nExpected lines:\n" + "\n".join(pattern)
            )

        matched_lines = output_lines[match_index : match_index + len(pattern)]
        replacement_lines = _build_context_replacement(hunk, matched_lines)
        output_lines[match_index : match_index + len(pattern)] = replacement_lines
        search_index = match_index + len(replacement_lines)

    return _render_file_lines(output_lines)


def _parse_unified_diff_lines(lines: list[str]) -> list[_UnifiedDiffHunk]:
    index = 0
    hunks: list[_UnifiedDiffHunk] = []

    while index < len(lines):
        line = lines[index]
        if not line:
            index += 1
            continue
        if line.startswith(_DIFF_HEADER_PREFIXES):
            index += 1
            continue

        match = _HUNK_HEADER_RE.match(line)
        if match is None:
            raise ApplyPatchError(f"Unsupported apply_patch diff line: {line}")

        old_count = int(match.group("old_count") or "1")
        new_count = int(match.group("new_count") or "1")
        index += 1

        hunk_lines: list[_DiffLine] = []
        while index < len(lines):
            next_line = lines[index]
            if _HUNK_HEADER_RE.match(next_line) or next_line.startswith(_DIFF_HEADER_PREFIXES):
                break
            if next_line == _NO_NEWLINE_MARKER:
                if not hunk_lines:
                    raise ApplyPatchError("apply_patch diff used a no-newline marker before any hunk line.")
                hunk_lines[-1].has_newline = False
                index += 1
                continue
            if next_line and next_line[0] in {" ", "+", "-"}:
                hunk_lines.append(_DiffLine(prefix=next_line[0], text=next_line[1:]))
                index += 1
                continue
            raise ApplyPatchError(f"Unsupported apply_patch hunk line: {next_line}")

        consumed_old = sum(1 for hunk_line in hunk_lines if hunk_line.prefix != "+")
        produced_new = sum(1 for hunk_line in hunk_lines if hunk_line.prefix != "-")
        if consumed_old != old_count:
            raise ApplyPatchError("apply_patch old-line count did not match the diff body.")
        if produced_new != new_count:
            raise ApplyPatchError("apply_patch new-line count did not match the diff body.")

        hunks.append(
            _UnifiedDiffHunk(
                old_start=int(match.group("old_start")),
                old_count=old_count,
                new_start=int(match.group("new_start")),
                new_count=new_count,
                lines=tuple(hunk_lines),
            )
        )

    return hunks


def _parse_context_diff_lines(lines: list[str]) -> list[_ContextDiffHunk]:
    index = 0
    hunks: list[_ContextDiffHunk] = []

    while index < len(lines):
        line = lines[index]
        if not line:
            index += 1
            continue
        if line.startswith(_DIFF_HEADER_PREFIXES):
            index += 1
            continue

        if line == _EMPTY_CONTEXT_HUNK:
            change_context = None
        elif line.startswith(_CONTEXT_HUNK_PREFIX):
            change_context = line[len(_CONTEXT_HUNK_PREFIX) :]
        else:
            raise ApplyPatchError(f"Unsupported apply_patch diff line: {line}")
        index += 1

        hunk_lines: list[_DiffLine] = []
        is_end_of_file = False
        while index < len(lines):
            next_line = lines[index]
            if _is_context_hunk_header(next_line) or next_line.startswith(_DIFF_HEADER_PREFIXES):
                break
            if next_line == _END_OF_FILE_MARKER:
                if not hunk_lines:
                    raise ApplyPatchError("apply_patch diff used an end-of-file marker before any hunk line.")
                is_end_of_file = True
                index += 1
                break
            if next_line == _NO_NEWLINE_MARKER:
                if not hunk_lines:
                    raise ApplyPatchError("apply_patch diff used a no-newline marker before any hunk line.")
                hunk_lines[-1].has_newline = False
                index += 1
                continue
            if next_line == "":
                hunk_lines.append(_DiffLine(prefix=" ", text=""))
                index += 1
                continue
            if next_line[0] in {" ", "+", "-"}:
                hunk_lines.append(_DiffLine(prefix=next_line[0], text=next_line[1:]))
                index += 1
                continue
            raise ApplyPatchError(f"Unsupported apply_patch hunk line: {next_line}")

        if not hunk_lines:
            raise ApplyPatchError("apply_patch context hunk did not contain any lines.")

        hunks.append(
            _ContextDiffHunk(change_context=change_context, lines=tuple(hunk_lines), is_end_of_file=is_end_of_file)
        )

    return hunks


def _resolve_hunk_source_index(hunk: _UnifiedDiffHunk) -> int:
    if hunk.old_count == 0:
        return hunk.old_start
    return max(hunk.old_start - 1, 0)


def _is_context_hunk_header(line: str) -> bool:
    return line == _EMPTY_CONTEXT_HUNK or line.startswith(_CONTEXT_HUNK_PREFIX)


def _build_context_replacement(hunk: _ContextDiffHunk, matched_lines: list[_FileLine]) -> list[_FileLine]:
    replacement_lines: list[_FileLine] = []
    matched_index = 0

    for diff_line in hunk.lines:
        if diff_line.prefix == " ":
            actual_line = _get_source_line(matched_lines, matched_index)
            _assert_line_matches(actual_line, diff_line.text)
            replacement_lines.append(actual_line)
            matched_index += 1
        elif diff_line.prefix == "-":
            actual_line = _get_source_line(matched_lines, matched_index)
            _assert_line_matches(actual_line, diff_line.text)
            matched_index += 1
        elif diff_line.prefix == "+":
            replacement_lines.append(_FileLine(text=diff_line.text, has_newline=diff_line.has_newline))
        else:  # pragma: no cover - defensive branch
            raise ApplyPatchError(f"Unsupported apply_patch context line prefix: {diff_line.prefix}")

    if matched_index != len(matched_lines):
        raise ApplyPatchError("apply_patch context hunk did not consume the expected source lines.")

    return replacement_lines


def _seek_sequence(lines: list[_FileLine], pattern: list[str], start: int, eof: bool) -> int | None:
    if not pattern:
        return len(lines) if eof else start
    if len(pattern) > len(lines):
        return None

    search_start = len(lines) - len(pattern) if eof and len(lines) >= len(pattern) else start
    last_index = len(lines) - len(pattern)
    for normalizer in (_identity, str.rstrip, str.strip, _normalise):
        for index in range(search_start, last_index + 1):
            if _matches_with(lines, pattern, index, normalizer):
                return index
    return None


def _matches_with(lines: list[_FileLine], pattern: list[str], start: int, normalizer: Callable[[str], str]) -> bool:
    if start + len(pattern) > len(lines):
        return False
    for offset, expected_text in enumerate(pattern):
        if normalizer(lines[start + offset].text) != normalizer(expected_text):
            return False
    return True


def _identity(value: str) -> str:
    return value


def _split_file_lines(text: str) -> list[_FileLine]:
    if not text:
        return []

    lines: list[_FileLine] = []
    for raw_line in text.splitlines(keepends=True):
        has_newline = raw_line.endswith("\n")
        stripped = raw_line[:-1] if has_newline else raw_line
        if stripped.endswith("\r"):
            stripped = stripped[:-1]
        lines.append(_FileLine(text=stripped, has_newline=has_newline))
    return lines


def _render_file_lines(lines: list[_FileLine]) -> str:
    return "".join(line.text + ("\n" if line.has_newline else "") for line in lines)


def _get_source_line(lines: list[_FileLine], index: int) -> _FileLine:
    try:
        return lines[index]
    except IndexError as exc:  # pragma: no cover - defensive branch
        raise ApplyPatchError("apply_patch expected more source lines than were available.") from exc


def _assert_line_matches(actual_line: _FileLine, expected_text: str) -> None:
    if _same_line(actual_line.text, expected_text):
        return
    raise ApplyPatchError(
        "apply_patch diff did not match the current file contents.\n"
        f"Expected: {expected_text!r}\n"
        f"Actual: {actual_line.text!r}"
    )


def _same_line(actual: str, expected: str) -> bool:
    if actual == expected:
        return True
    if actual.rstrip() == expected.rstrip():
        return True
    if actual.strip() == expected.strip():
        return True
    return _normalise(actual) == _normalise(expected)


def _normalise(value: str) -> str:
    mapping = {
        "\u2010": "-",
        "\u2011": "-",
        "\u2012": "-",
        "\u2013": "-",
        "\u2014": "-",
        "\u2015": "-",
        "\u2212": "-",
        "\u2018": "'",
        "\u2019": "'",
        "\u201a": "'",
        "\u201b": "'",
        "\u201c": '"',
        "\u201d": '"',
        "\u201e": '"',
        "\u201f": '"',
        "\u00a0": " ",
        "\u2002": " ",
        "\u2003": " ",
        "\u2004": " ",
        "\u2005": " ",
        "\u2006": " ",
        "\u2007": " ",
        "\u2008": " ",
        "\u2009": " ",
        "\u200a": " ",
        "\u202f": " ",
        "\u205f": " ",
        "\u3000": " ",
    }
    return "".join(mapping.get(ch, ch) for ch in value.strip())



# Generated by ariadne-codegen
# Source: exponent/core/graphql/operations

from typing import Literal
from uuid import UUID

from pydantic import Field

from .base_model import BaseModel


class CreateCloudChatFromRepository(BaseModel):
    create_chat_with_resource_config: (
        "CreateCloudChatFromRepositoryCreateChatWithResourceConfig"
    ) = Field(alias="createChatWithResourceConfig")


class CreateCloudChatFromRepositoryCreateChatWithResourceConfig(BaseModel):
    typename__: Literal["Chat"] = Field(alias="__typename")
    chat_uuid: UUID = Field(alias="chatUuid")


CreateCloudChatFromRepository.model_rebuild()



"""indent-ssh-helper: tunnels SSH over a WebRTC data channel."""

import asyncio
import json
import os
import sys
import uuid
from pathlib import Path

import websockets
from aiortc import RTCConfiguration, RTCIceServer, RTCPeerConnection, RTCSessionDescription
from aiortc.sdp import candidate_from_sdp


def _log(msg: str, **fields: object) -> None:
    parts = [msg] + [f"{k}={v}" for k, v in fields.items()]
    sys.stderr.write(" ".join(parts) + "\n")
    sys.stderr.flush()


def _read_api_key() -> str:
    key = os.environ.get("INDENT_API_KEY")
    if key:
        return key

    config_path = Path.home() / ".config" / "indent" / "config.json"
    try:
        config = json.loads(config_path.read_text())
    except (OSError, json.JSONDecodeError) as e:
        raise SystemExit(f"cannot read {config_path}: {e} (set INDENT_API_KEY env var as an alternative)")

    env = os.environ.get("INDENT_ENV")
    if env:
        extra = config.get("extra_exponent_api_keys", {})
        if env in extra:
            return extra[env]
        raise SystemExit(f"INDENT_ENV={env!r} but no matching key in extra_exponent_api_keys in {config_path}")

    api_key = config.get("exponent_api_key")
    if not api_key:
        raise SystemExit(f"exponent_api_key not found in {config_path}")
    return api_key


async def _run(chat_uuid: str, api_key: str, ws_url: str) -> None:
    _log("connecting", chat_uuid=chat_uuid, ws_url=ws_url)

    async with websockets.connect(ws_url, open_timeout=10) as ws:
        await ws.send(json.dumps({"type": "auth", "data": {"api_key": api_key}}))
        auth = json.loads(await ws.recv())
        if auth.get("type") != "auth_success":
            raise SystemExit(f"auth failed: {auth.get('reason')}")
        _log("authenticated", role=auth.get("role"))

        ice_configs = auth.get("ice_servers", [])
        ice_servers = []
        for c in ice_configs:
            ice_servers.append(
                RTCIceServer(
                    urls=c.get("urls", []),
                    username=c.get("username", ""),
                    credential=c.get("credential", ""),
                )
            )

        pc = RTCPeerConnection(RTCConfiguration(iceServers=ice_servers))

        done = asyncio.Event()
        loop = asyncio.get_running_loop()
        background_tasks: set[asyncio.Task] = set()

        @pc.on("connectionstatechange")
        async def on_state() -> None:
            _log("peer connection state", state=pc.connectionState)
            if pc.connectionState in ("failed", "closed"):
                done.set()

        # Negotiated SCTP channel (must match remote side)
        pc.createDataChannel("_sctp", negotiated=True, id=0)

        ssh_dc = pc.createDataChannel("ssh", ordered=True)

        @ssh_dc.on("open")
        def on_open() -> None:
            _log("ssh data channel open")

            async def pump_stdin() -> None:
                reader = asyncio.StreamReader()
                await loop.connect_read_pipe(lambda: asyncio.StreamReaderProtocol(reader), sys.stdin.buffer)
                try:
                    while True:
                        chunk = await reader.read(16384)
                        if not chunk:
                            break
                        ssh_dc.send(chunk)
                except Exception:
                    pass
                done.set()

            task = loop.create_task(pump_stdin())
            background_tasks.add(task)
            task.add_done_callback(background_tasks.discard)

        @ssh_dc.on("message")
        def on_message(msg: bytes | str) -> None:
            data = msg if isinstance(msg, bytes) else msg.encode()
            try:
                sys.stdout.buffer.write(data)
                sys.stdout.buffer.flush()
            except BrokenPipeError:
                done.set()

        @ssh_dc.on("close")
        def on_close() -> None:
            _log("ssh data channel closed")
            done.set()

        offer = await pc.createOffer()
        await pc.setLocalDescription(offer)

        local_desc = pc.localDescription
        offer_payload = json.dumps({"type": "offer", "sdp": local_desc.sdp})
        session_id = str(uuid.uuid4())

        await ws.send(
            json.dumps(
                {
                    "type": "offer",
                    "session_id": session_id,
                    "target_port": 22,
                    "payload": offer_payload,
                }
            )
        )
        _log("offer sent", session_id=session_id)

        async def handle_signaling() -> None:
            try:
                async for raw in ws:
                    msg = json.loads(raw)
                    match msg.get("type"):
                        case "session_ok":
                            _log("session accepted", session_id=msg.get("session_id"))
                        case "signal":
                            kind = msg.get("kind")
                            payload = json.loads(msg.get("payload", "{}"))
                            if kind == "ANSWER":
                                answer = RTCSessionDescription(sdp=payload["sdp"], type="answer")
                                await pc.setRemoteDescription(answer)
                                _log("remote description set (answer)")
                            elif kind == "ICE_CANDIDATE":
                                candidate_str = payload.get("candidate", "")
                                if candidate_str:
                                    candidate = candidate_from_sdp(candidate_str)
                                    candidate.sdpMid = payload.get("sdpMid", "0")
                                    candidate.sdpMLineIndex = payload.get("sdpMLineIndex", 0)
                                    await pc.addIceCandidate(candidate)
                        case "error":
                            _log("signaling error", error=msg.get("error"), session_id=msg.get("session_id"))
                            done.set()
                            return
                        case "ping":
                            await ws.send(json.dumps({"type": "pong"}))
            except websockets.ConnectionClosed:
                pass
            done.set()

        signaling_task = asyncio.ensure_future(handle_signaling())

        try:
            await done.wait()
        except asyncio.CancelledError:
            pass
        finally:
            _log("exiting")
            signaling_task.cancel()
            try:
                await signaling_task
            except (asyncio.CancelledError, Exception):
                pass
            await pc.close()


def main() -> None:
    if len(sys.argv) < 2:
        sys.stderr.write("Usage: indent-ssh-helper <uuid>.indent\n")
        sys.exit(1)

    host = sys.argv[1]
    chat_uuid = host.removesuffix(".indent")
    try:
        uuid.UUID(chat_uuid)
    except ValueError:
        sys.stderr.write(f'Error: "{chat_uuid}" is not a valid chat UUID\n')
        sys.exit(1)

    api_key = _read_api_key()

    ws_url = os.environ.get("INDENT_WS_URL")
    if not ws_url:
        ws_url = f"wss://ws-api.indent.com/api/ws/webrtc/{chat_uuid}"
    else:
        ws_url = f"{ws_url.rstrip('/')}/webrtc/{chat_uuid}"

    try:
        asyncio.run(_run(chat_uuid, api_key, ws_url))
    except KeyboardInterrupt:
        pass



# Generated by ariadne-codegen
# Source: ../../schema.graphql

from typing import Any, Optional
from uuid import UUID

from pydantic import Field

from .base_model import BaseModel
from .enums import (
    AgentSubtype,
    ChatMode,
    ModelName,
    ReasoningLevel,
    SandboxProvider,
    ToolPermissionStatus,
)


class AcceptPlanInput(BaseModel):
    tool_use_id: str = Field(alias="toolUseId")
    clear_context: bool = Field(alias="clearContext")


class AskUserQuestionAnswerInput(BaseModel):
    question: str
    selected_options: list[str] = Field(alias="selectedOptions")
    custom_answer: Optional[str] = Field(alias="customAnswer", default=None)


class AskUserQuestionInput(BaseModel):
    answers: list["AskUserQuestionAnswerInput"]
    skip: bool = False
    tool_use_id: Optional[str] = Field(alias="toolUseId", default=None)


class AttachmentInput(BaseModel):
    file_attachment: Optional["FileAttachmentInput"] = Field(
        alias="fileAttachment", default=None
    )
    url_attachment: Optional["URLAttachmentInput"] = Field(
        alias="urlAttachment", default=None
    )
    table_schema_attachment: Optional["TableSchemaAttachmentInput"] = Field(
        alias="tableSchemaAttachment", default=None
    )
    prompt_attachment: Optional["PromptAttachmentInput"] = Field(
        alias="promptAttachment", default=None
    )
    sql_attachment: Optional["SQLAttachmentInput"] = Field(
        alias="sqlAttachment", default=None
    )
    skill_attachment: Optional["SkillAttachmentInput"] = Field(
        alias="skillAttachment", default=None
    )


class ChatConfigInput(BaseModel):
    chat_uuid: UUID = Field(alias="chatUuid")
    model: Optional[ModelName] = None
    require_confirmation: Optional[bool] = Field(
        alias="requireConfirmation", default=False
    )
    depth_limit: int = Field(alias="depthLimit", default=6)
    reasoning_level: Optional[ReasoningLevel] = Field(
        alias="reasoningLevel", default=None
    )
    internal_settings: Optional[Any] = Field(alias="internalSettings", default=None)


class ChatInput(BaseModel):
    prompt_input: Optional["PromptInput"] = Field(alias="promptInput", default=None)
    client_sourced_event_uuid: Optional[UUID] = Field(
        alias="clientSourcedEventUuid", default=None
    )
    permission_response: Optional["PermissionResponseInput"] = Field(
        alias="permissionResponse", default=None
    )
    accept_plan_input: Optional["AcceptPlanInput"] = Field(
        alias="acceptPlanInput", default=None
    )
    enter_plan_mode: Optional[bool] = Field(alias="enterPlanMode", default=None)
    exit_plan_mode_without_plan: Optional[bool] = Field(
        alias="exitPlanModeWithoutPlan", default=None
    )
    environment_variable_response: Optional["EnvironmentVariableResponseInput"] = Field(
        alias="environmentVariableResponse", default=None
    )
    ask_user_question: Optional["AskUserQuestionInput"] = Field(
        alias="askUserQuestion", default=None
    )
    timezone: Optional[str] = None
    mcp_server_configs: Optional["McpServerConfigsInput"] = Field(
        alias="mcpServerConfigs", default=None
    )
    database_resource_configs: Optional[list["DatabaseResourceConfigInput"]] = Field(
        alias="databaseResourceConfigs", default=None
    )
    repository_resource_configs: Optional[list["RepositoryResourceConfigInput"]] = (
        Field(alias="repositoryResourceConfigs", default=None)
    )
    include_datadog_tools: Optional[bool] = Field(
        alias="includeDatadogTools", default=None
    )
    include_sentry_tools: Optional[bool] = Field(
        alias="includeSentryTools", default=None
    )
    include_aws_tools: Optional[bool] = Field(alias="includeAwsTools", default=None)
    enabled_skill_ids: Optional[list[str]] = Field(
        alias="enabledSkillIds", default=None
    )


class ChatResourceConfigInput(BaseModel):
    mode: ChatMode
    primary_repository: Optional["RepositoryResourceConfigInput"] = Field(
        alias="primaryRepository", default=None
    )
    additional_repositories: Optional[list["RepositoryResourceConfigInput"]] = Field(
        alias="additionalRepositories", default=None
    )
    database: Optional["DatabaseResourceConfigInput"] = None
    mcp_server_configs: Optional["McpServerConfigsInput"] = Field(
        alias="mcpServerConfigs", default=None
    )
    skill_configs: Optional["SkillConfigsInput"] = Field(
        alias="skillConfigs", default=None
    )
    sandbox_provider: Optional[SandboxProvider] = Field(
        alias="sandboxProvider", default=None
    )
    cloud_agent_subtype: Optional[AgentSubtype] = Field(
        alias="cloudAgentSubtype", default=None
    )


class DatabaseResourceConfigInput(BaseModel):
    database_config_uuid: UUID = Field(alias="databaseConfigUuid")


class EnvironmentVariableResponseEntry(BaseModel):
    key: str
    value: str


class EnvironmentVariableResponseInput(BaseModel):
    variables: list["EnvironmentVariableResponseEntry"]
    tool_use_id: str = Field(alias="toolUseId")
    skipped_keys: list[str] = Field(alias="skippedKeys", default_factory=lambda: [])


class FileAttachmentInput(BaseModel):
    file: "FileInput"
    content: str


class FileInput(BaseModel):
    file_path: str = Field(alias="filePath")
    working_directory: str = Field(alias="workingDirectory")


class McpServerConfigInput(BaseModel):
    server_name: str = Field(alias="serverName")
    enabled_tools: list[str] = Field(alias="enabledTools")


class McpServerConfigsInput(BaseModel):
    all_servers: Optional[bool] = Field(alias="allServers", default=None)
    server_configs: Optional[list["McpServerConfigInput"]] = Field(
        alias="serverConfigs", default=None
    )


class PermissionResponseInput(BaseModel):
    permission_status: ToolPermissionStatus = Field(alias="permissionStatus")
    tool_use_id: str = Field(alias="toolUseId")


class PromptAttachmentInput(BaseModel):
    prompt_name: str = Field(alias="promptName")
    prompt_content: str = Field(alias="promptContent")


class PromptInput(BaseModel):
    message: str
    attachments: list["AttachmentInput"]


class RepositoryResourceConfigInput(BaseModel):
    repository_uuid: UUID = Field(alias="repositoryUuid")


class SQLAttachmentInput(BaseModel):
    query_content: str = Field(alias="queryContent")


class SkillAttachmentInput(BaseModel):
    skill_name: str = Field(alias="skillName")


class SkillConfigInput(BaseModel):
    skill_id: str = Field(alias="skillId")


class SkillConfigsInput(BaseModel):
    skill_configs: list["SkillConfigInput"] = Field(alias="skillConfigs")


class TableSchemaAttachmentInput(BaseModel):
    table_name: str = Field(alias="tableName")
    table_schema: Any = Field(alias="tableSchema")


class URLAttachmentInput(BaseModel):
    url: str
    content: str


AskUserQuestionInput.model_rebuild()
AttachmentInput.model_rebuild()
ChatInput.model_rebuild()
ChatResourceConfigInput.model_rebuild()
EnvironmentVariableResponseInput.model_rebuild()
FileAttachmentInput.model_rebuild()
McpServerConfigsInput.model_rebuild()
PromptInput.model_rebuild()
SkillConfigsInput.model_rebuild()



# Generated by ariadne-codegen
# Source: exponent/core/graphql/operations

from typing import Literal
from uuid import UUID

from pydantic import Field

from .base_model import BaseModel


class StartChatTurn(BaseModel):
    start_chat_turn: "StartChatTurnStartChatTurn" = Field(alias="startChatTurn")


class StartChatTurnStartChatTurn(BaseModel):
    typename__: Literal["Chat"] = Field(alias="__typename")
    chat_uuid: UUID = Field(alias="chatUuid")


StartChatTurn.model_rebuild()



# Generated by ariadne-codegen

from .base_client import AsyncBaseClient
from .base_model import BaseModel, Upload
from .chats import (
    Chats,
    ChatsOrganizationChatsPage,
    ChatsOrganizationChatsPageChats,
    ChatsOrganizationChatsPageChatsPrStatusNotFoundError,
    ChatsOrganizationChatsPageChatsPrStatusPRInfo,
)
from .client import IndentGraphQLClient
from .create_cloud_chat_from_repository import (
    CreateCloudChatFromRepository,
    CreateCloudChatFromRepositoryCreateChatWithResourceConfig,
)
from .current_user import CurrentUser, CurrentUserCurrentUser
from .enums import (
    AgentSubtype,
    ChatMode,
    ModelName,
    PRStatus,
    ReasoningLevel,
    SandboxProvider,
    ToolPermissionStatus,
)
from .github_repositories import (
    GithubRepositories,
    GithubRepositoriesGithubRepositories,
    GithubRepositoriesGithubRepositoriesRepositories,
)
from .halt_chat_stream import (
    HaltChatStream,
    HaltChatStreamHaltChatStreamChat,
    HaltChatStreamHaltChatStreamRemoteExecutionError,
)
from .input_types import (
    AcceptPlanInput,
    AskUserQuestionAnswerInput,
    AskUserQuestionInput,
    AttachmentInput,
    ChatConfigInput,
    ChatInput,
    ChatResourceConfigInput,
    DatabaseResourceConfigInput,
    EnvironmentVariableResponseEntry,
    EnvironmentVariableResponseInput,
    FileAttachmentInput,
    FileInput,
    McpServerConfigInput,
    McpServerConfigsInput,
    PermissionResponseInput,
    PromptAttachmentInput,
    PromptInput,
    RepositoryResourceConfigInput,
    SkillAttachmentInput,
    SkillConfigInput,
    SkillConfigsInput,
    SQLAttachmentInput,
    TableSchemaAttachmentInput,
    URLAttachmentInput,
)
from .refresh_api_key import RefreshApiKey, RefreshApiKeyRefreshApiKey
from .set_login_complete import SetLoginComplete, SetLoginCompleteSetLoginComplete
from .start_chat_turn import StartChatTurn, StartChatTurnStartChatTurn

__all__ = [
    "AcceptPlanInput",
    "AgentSubtype",
    "AskUserQuestionAnswerInput",
    "AskUserQuestionInput",
    "AsyncBaseClient",
    "AttachmentInput",
    "BaseModel",
    "ChatConfigInput",
    "ChatInput",
    "ChatMode",
    "ChatResourceConfigInput",
    "Chats",
    "ChatsOrganizationChatsPage",
    "ChatsOrganizationChatsPageChats",
    "ChatsOrganizationChatsPageChatsPrStatusNotFoundError",
    "ChatsOrganizationChatsPageChatsPrStatusPRInfo",
    "CreateCloudChatFromRepository",
    "CreateCloudChatFromRepositoryCreateChatWithResourceConfig",
    "CurrentUser",
    "CurrentUserCurrentUser",
    "DatabaseResourceConfigInput",
    "EnvironmentVariableResponseEntry",
    "EnvironmentVariableResponseInput",
    "FileAttachmentInput",
    "FileInput",
    "GithubRepositories",
    "GithubRepositoriesGithubRepositories",
    "GithubRepositoriesGithubRepositoriesRepositories",
    "HaltChatStream",
    "HaltChatStreamHaltChatStreamChat",
    "HaltChatStreamHaltChatStreamRemoteExecutionError",
    "IndentGraphQLClient",
    "McpServerConfigInput",
    "McpServerConfigsInput",
    "ModelName",
    "PRStatus",
    "PermissionResponseInput",
    "PromptAttachmentInput",
    "PromptInput",
    "ReasoningLevel",
    "RefreshApiKey",
    "RefreshApiKeyRefreshApiKey",
    "RepositoryResourceConfigInput",
    "SQLAttachmentInput",
    "SandboxProvider",
    "SetLoginComplete",
    "SetLoginCompleteSetLoginComplete",
    "SkillAttachmentInput",
    "SkillConfigInput",
    "SkillConfigsInput",
    "StartChatTurn",
    "StartChatTurnStartChatTurn",
    "TableSchemaAttachmentInput",
    "ToolPermissionStatus",
    "URLAttachmentInput",
    "Upload",
]



from __future__ import annotations

import os
import secrets

from exponent.core.config import get_chat_artifacts_dir

BASH_RESULTS_DIR = "bash_results"
QUERY_RESULTS_DIR = "query_results"
USER_UPLOADS_DIR = "user_uploads"
ARTIFACTS_DIR = "artifacts"
ARTIFACTS_HISTORY_DIR = "artifacts/history"
CONVERSATION_HISTORY_FILENAME = "conversation_history.jsonl"


def generate_bash_id() -> str:
    alphabet = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
    return "".join(secrets.choice(alphabet) for _ in range(8))


def _paths(chat_uuid: str, relative: str) -> tuple[str, str]:
    return os.path.join(get_chat_artifacts_dir(str(chat_uuid)), relative), relative


def bash_result_path(chat_uuid: str, bash_id: str, epoch_ms: int) -> tuple[str, str]:
    return _paths(chat_uuid, f"{BASH_RESULTS_DIR}/{bash_id}_{epoch_ms}.txt")


def query_result_path(chat_uuid: str, query_id: str, epoch_ms: int, ext: str = "parquet") -> tuple[str, str]:
    return _paths(chat_uuid, f"{QUERY_RESULTS_DIR}/{query_id}_{epoch_ms}.{ext}")


def user_upload_path(chat_uuid: str, sanitized_filename: str) -> tuple[str, str]:
    return _paths(chat_uuid, f"{USER_UPLOADS_DIR}/{sanitized_filename}")


def artifact_path(chat_uuid: str, filename: str) -> tuple[str, str]:
    return _paths(chat_uuid, f"{ARTIFACTS_DIR}/{filename}")


def conversation_history_path(chat_uuid: str) -> tuple[str, str]:
    return _paths(chat_uuid, CONVERSATION_HISTORY_FILENAME)


def s3_artifact_key(filename: str) -> str:
    return f"{ARTIFACTS_DIR}/{filename}"


def s3_artifact_history_key(filename: str, version: int) -> str:
    return f"{ARTIFACTS_HISTORY_DIR}/{filename}/{version}"


def s3_query_result_csv_key(query_id: str, epoch_ms: int) -> str:
    return f"{QUERY_RESULTS_DIR}/{query_id}_{epoch_ms}.csv"


def s3_key_to_local_path(s3_key: str, chat_uuid: str) -> str:
    return os.path.join(get_chat_artifacts_dir(chat_uuid), s3_key)


def local_path_to_s3_key(local_path: str, chat_uuid: str) -> str:
    chat_dir = get_chat_artifacts_dir(chat_uuid)
    return os.path.relpath(local_path, chat_dir)


_SUBDIRS = [
    BASH_RESULTS_DIR,
    QUERY_RESULTS_DIR,
    USER_UPLOADS_DIR,
    ARTIFACTS_HISTORY_DIR,
]


def ensure_chat_directories(chat_uuid: str) -> None:
    chat_dir = get_chat_artifacts_dir(chat_uuid)
    for subdir in _SUBDIRS:
        os.makedirs(os.path.join(chat_dir, subdir), exist_ok=True)



# Generated by ariadne-codegen
# Source: exponent/core/graphql/operations

from typing import Literal

from pydantic import Field

from .base_model import BaseModel


class SetLoginComplete(BaseModel):
    set_login_complete: "SetLoginCompleteSetLoginComplete" = Field(
        alias="setLoginComplete"
    )


class SetLoginCompleteSetLoginComplete(BaseModel):
    typename__: Literal["User"] = Field(alias="__typename")
    user_api_key: str = Field(alias="userApiKey")


SetLoginComplete.model_rebuild()



# Generated by ariadne-codegen
# Source: exponent/core/graphql/operations

from uuid import UUID

from pydantic import Field

from .base_model import BaseModel


class CurrentUser(BaseModel):
    current_user: "CurrentUserCurrentUser" = Field(alias="currentUser")


class CurrentUserCurrentUser(BaseModel):
    user_uuid: UUID = Field(alias="userUuid")


CurrentUser.model_rebuild()



import asyncio
import json
import logging
import os
import struct
from collections.abc import Awaitable, Callable

from vt100wasm import Screen

from exponent.core.remote_execution.terminal_types import (
    TerminalMessage,
    TerminalOutput,
    TerminalResetSessions,
    TerminalStatus,
)

InputCallback = Callable[[str, str], Awaitable[None]]
ResizeCallback = Callable[[str, int, int], Awaitable[None]]
StartTerminalCallback = Callable[[str, int, int], Awaitable[None]]


logger = logging.getLogger(__name__)

TAG_SYNC_COMPLETE = 0
TAG_TERMINAL_STATE = 1
TAG_TERMINAL_SNAPSHOT = 2
TAG_TERMINAL_STREAM = 3
TAG_START_STREAMING = 4
TAG_STOP_STREAMING = 5
TAG_BG_TERMINAL_STATE = 6
TAG_TERMINAL_INPUT = 7
TAG_TERMINAL_RESIZE = 8
TAG_START_TERMINAL = 9
TAG_TERMINAL_CELL_DIFF = 11
TAG_TERMINAL_CELLS = 12
TAG_TERMINAL_SCROLL = 13
TAG_TERMINAL_CLEAR = 14

HEADER_SIZE = 5
MAX_PAYLOAD_SIZE = 10 * 1024 * 1024
DEFAULT_ROWS = 24
DEFAULT_COLS = 80
BG_TAIL_INTERVAL = 0.25
OUTPUT_DEBOUNCE_S = 1 / 250

SCROLLBACK_BYTES = 1_048_576
CELL_RECORD_SIZE = 16

SYNC_UPDATE_BEGIN = "\x1b[?2026h"
SYNC_UPDATE_END = "\x1b[?2026l"


def encode_tlv(tag: int, payload: bytes) -> bytes:
    return struct.pack("!BI", tag, len(payload)) + payload


def decode_tlv_header(data: bytes) -> tuple[int, int]:
    tag, length = struct.unpack("!BI", data[:HEADER_SIZE])
    return tag, length


def _scrollback_rows(cols: int) -> int:
    return max(1, SCROLLBACK_BYTES // (max(1, cols) * CELL_RECORD_SIZE))


class TerminalState:
    __slots__ = ("cols", "id", "name", "rows", "running", "screen")

    def __init__(self, id: str, name: str | None, rows: int = DEFAULT_ROWS, cols: int = DEFAULT_COLS) -> None:
        self.id = id
        self.name = name
        self.running = True
        self.screen: Screen | None = None
        self.rows = rows
        self.cols = cols

    def state_label(self) -> str:
        return "running" if self.running else "completed"

    def to_state_payload(self) -> bytes:
        return json.dumps({"id": self.id, "name": self.name, "state": self.state_label()}).encode()

    def apply_output(self, data: bytes) -> None:
        if self.screen is None:
            self.screen = Screen(self.rows, self.cols, _scrollback_rows(self.cols))
        self.screen.process(data)

    def snapshot_payload(self) -> bytes:
        if self.screen is None:
            return b""
        return self.screen.snapshot()


class BackgroundTerminalState:
    __slots__ = ("_file_offset", "cols", "command", "exit_code", "id", "output_file", "rows", "running", "screen")

    def __init__(self, id: str, command: str, output_file: str) -> None:
        self.id = id
        self.command = command
        self.output_file = output_file
        self.running = True
        self.exit_code: int | None = None
        self.screen: Screen | None = None
        self.rows = DEFAULT_ROWS
        self.cols = DEFAULT_COLS
        self._file_offset = 0

    def state_label(self) -> str:
        return "running" if self.running else "completed"

    def to_state_payload(self) -> bytes:
        d: dict[str, object] = {
            "id": self.id,
            "command": self.command,
            "output_file": self.output_file,
            "state": self.state_label(),
        }
        if self.exit_code is not None:
            d["exit_code"] = self.exit_code
        return json.dumps(d).encode()

    def read_new_output(self) -> bytes:
        try:
            with open(self.output_file, "rb") as f:
                f.seek(self._file_offset)
                data = f.read()
            if data:
                self._file_offset += len(data)
                if self.screen is None:
                    self.screen = Screen(self.rows, self.cols, _scrollback_rows(self.cols))
                self.screen.process(data)
            return data
        except FileNotFoundError:
            return b""

    def snapshot_payload(self) -> bytes:
        if self.screen is None:
            return b""
        return self.screen.snapshot()


class ClientConnection:
    __slots__ = ("_subscriptions", "_write_lock", "_writer")

    def __init__(self, writer: asyncio.StreamWriter) -> None:
        self._writer = writer
        self._subscriptions: set[str] = set()
        self._write_lock = asyncio.Lock()

    async def send(self, data: bytes) -> bool:
        try:
            async with self._write_lock:
                self._writer.write(data)
                await self._writer.drain()
            return True
        except (ConnectionError, OSError):
            return False

    def subscribe(self, terminal_id: str) -> None:
        self._subscriptions.add(terminal_id)

    def unsubscribe(self, terminal_id: str) -> None:
        self._subscriptions.discard(terminal_id)

    def is_subscribed(self, terminal_id: str) -> bool:
        return terminal_id in self._subscriptions

    def close(self) -> None:
        self._writer.close()


class TerminalControllerServer:
    def __init__(
        self,
        socket_path: str,
        on_input: InputCallback | None = None,
        on_resize: ResizeCallback | None = None,
        on_start_terminal: StartTerminalCallback | None = None,
    ) -> None:
        self._socket_path = socket_path
        self._terminals: dict[str, TerminalState] = {}
        self._bg_terminals: dict[str, BackgroundTerminalState] = {}
        self._bg_tail_tasks: dict[str, asyncio.Task[None]] = {}
        self._clients: set[ClientConnection] = set()
        self._server: asyncio.Server | None = None
        self._on_input = on_input
        self._on_resize = on_resize
        self._on_start_terminal = on_start_terminal
        self._pending_output: dict[str, bool] = {}
        self._flush_handles: dict[str, asyncio.TimerHandle] = {}
        self._in_sync_update: set[str] = set()

    async def start(self) -> None:
        try:
            os.unlink(self._socket_path)
        except FileNotFoundError:
            # Socket file does not exist yet; this is expected on first start.
            logger.debug("Socket file %s did not exist before start; nothing to remove", self._socket_path)
        self._server = await asyncio.start_unix_server(self._handle_client, path=self._socket_path)
        os.chmod(self._socket_path, 0o666)
        logger.info("Terminal controller listening on %s", self._socket_path)

    async def stop(self) -> None:
        for handle in self._flush_handles.values():
            handle.cancel()
        self._flush_handles.clear()
        self._pending_output.clear()
        self._in_sync_update.clear()
        for task in self._bg_tail_tasks.values():
            task.cancel()
        if self._bg_tail_tasks:
            await asyncio.gather(*self._bg_tail_tasks.values(), return_exceptions=True)
        self._bg_tail_tasks.clear()
        if self._server:
            self._server.close()
            await self._server.wait_closed()
        for client in list(self._clients):
            client.close()
        self._clients.clear()
        try:
            os.unlink(self._socket_path)
        except FileNotFoundError:
            pass  # Socket already removed during shutdown; expected.

    async def handle_terminal_message(self, message: TerminalMessage) -> None:
        await self._handle_terminal_message(message)

    async def _handle_terminal_message(self, message: TerminalMessage) -> None:
        if isinstance(message, TerminalResetSessions):
            self._terminals.clear()
            return

        if isinstance(message, TerminalStatus):
            await self._handle_status(message)
        elif isinstance(message, TerminalOutput):
            await self._handle_output(message)

    async def _handle_status(self, status: TerminalStatus) -> None:
        terminal = self._terminals.get(status.session_id)
        if status.status == "started":
            terminal = TerminalState(id=status.session_id, name=status.name)
            self._terminals[status.session_id] = terminal
        elif terminal is not None:
            terminal.running = False

        if terminal is not None:
            msg = encode_tlv(TAG_TERMINAL_STATE, terminal.to_state_payload())
            await self._broadcast(msg)

    async def _handle_output(self, output: TerminalOutput) -> None:
        terminal = self._terminals.get(output.session_id)
        if terminal is None:
            return

        data = output.data.encode("utf-8")
        terminal.apply_output(data)

        sid = output.session_id
        self._pending_output[sid] = True

        raw = output.data
        begin_pos = raw.rfind(SYNC_UPDATE_BEGIN)
        end_pos = raw.rfind(SYNC_UPDATE_END)
        if begin_pos >= 0 and begin_pos > end_pos:
            self._in_sync_update.add(sid)
            handle = self._flush_handles.pop(sid, None)
            if handle is not None:
                handle.cancel()
            return
        elif end_pos >= 0 and end_pos > begin_pos:
            self._in_sync_update.discard(sid)

        if sid in self._in_sync_update:
            return

        if sid not in self._flush_handles:
            loop = asyncio.get_running_loop()
            self._flush_handles[sid] = loop.call_later(
                OUTPUT_DEBOUNCE_S, lambda s=sid: asyncio.ensure_future(self._flush_output(s))
            )

    async def _flush_output(self, terminal_id: str) -> None:
        self._flush_handles.pop(terminal_id, None)
        if not self._pending_output.pop(terminal_id, False):
            return
        terminal = self._terminals.get(terminal_id)
        if terminal is None or terminal.screen is None:
            return
        sb = terminal.screen.scrollback()
        id_prefix = json.dumps({"id": terminal_id, "sb": sb}).encode() + b"\x00"
        if sb > 0:
            msg = encode_tlv(TAG_TERMINAL_CELLS, id_prefix + terminal.screen.cells())
        else:
            diff = terminal.screen.cells_diff()
            msg = encode_tlv(TAG_TERMINAL_CELL_DIFF, id_prefix + diff)
        await self._broadcast_to_subscribers(terminal_id, msg)

    async def handle_bg_process_started(self, process_id: str, command: str, output_file: str) -> None:
        bg = BackgroundTerminalState(id=process_id, command=command, output_file=output_file)
        self._bg_terminals[process_id] = bg
        msg = encode_tlv(TAG_BG_TERMINAL_STATE, bg.to_state_payload())
        await self._broadcast(msg)
        self._bg_tail_tasks[process_id] = asyncio.create_task(self._tail_bg_output(process_id))

    async def handle_bg_process_completed(self, process_id: str, exit_code: int) -> None:
        bg = self._bg_terminals.get(process_id)
        if bg is None:
            return
        task = self._bg_tail_tasks.pop(process_id, None)
        if task is not None:
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass  # Expected after task.cancel(); proceed to read remaining output.
        bg.read_new_output()
        bg.running = False
        bg.exit_code = exit_code
        msg = encode_tlv(TAG_BG_TERMINAL_STATE, bg.to_state_payload())
        await self._broadcast(msg)

    async def _tail_bg_output(self, process_id: str) -> None:
        try:
            while True:
                bg = self._bg_terminals.get(process_id)
                if bg is None or not bg.running:
                    return
                data = bg.read_new_output()
                if data:
                    msg = self._build_bg_snapshot_msg(process_id, bg)
                    await self._broadcast_to_subscribers(process_id, msg)
                await asyncio.sleep(BG_TAIL_INTERVAL)
        except asyncio.CancelledError:
            bg = self._bg_terminals.get(process_id)
            if bg is not None:
                data = bg.read_new_output()
                if data:
                    msg = self._build_bg_snapshot_msg(process_id, bg)
                    await self._broadcast_to_subscribers(process_id, msg)

    @staticmethod
    def _build_bg_snapshot_msg(terminal_id: str, bg: BackgroundTerminalState) -> bytes:
        id_prefix = json.dumps({"id": terminal_id}).encode() + b"\x00"
        return encode_tlv(TAG_TERMINAL_SNAPSHOT, id_prefix + bg.snapshot_payload())

    async def _broadcast(self, data: bytes) -> None:
        dead: list[ClientConnection] = []
        for client in self._clients:
            if not await client.send(data):
                dead.append(client)
        for client in dead:
            self._clients.discard(client)

    async def _broadcast_to_subscribers(self, terminal_id: str, data: bytes) -> None:
        dead: list[ClientConnection] = []
        for client in self._clients:
            if client.is_subscribed(terminal_id):
                if not await client.send(data):
                    dead.append(client)
        for client in dead:
            self._clients.discard(client)

    async def _handle_client(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:
        client = ClientConnection(writer)
        self._clients.add(client)
        logger.info("Terminal controller client connected, total=%d", len(self._clients))

        try:
            await self._send_initial_state(client)
            await self._read_client_messages(client, reader)
        except (ConnectionError, OSError, asyncio.IncompleteReadError):
            pass  # Client disconnected; cleanup handled in finally block.
        finally:
            self._clients.discard(client)
            client.close()
            logger.info("Terminal controller client disconnected, total=%d", len(self._clients))

    async def _send_initial_state(self, client: ClientConnection) -> None:
        for terminal in self._terminals.values():
            msg = encode_tlv(TAG_TERMINAL_STATE, terminal.to_state_payload())
            if not await client.send(msg):
                return
        for bg in self._bg_terminals.values():
            msg = encode_tlv(TAG_BG_TERMINAL_STATE, bg.to_state_payload())
            if not await client.send(msg):
                return
        await client.send(encode_tlv(TAG_SYNC_COMPLETE, b""))

    async def _read_client_messages(self, client: ClientConnection, reader: asyncio.StreamReader) -> None:
        while True:
            header = await reader.readexactly(HEADER_SIZE)
            tag, length = decode_tlv_header(header)
            if length > MAX_PAYLOAD_SIZE:
                logger.warning("Client sent oversized TLV payload (%d bytes), disconnecting", length)
                return
            payload = await reader.readexactly(length) if length > 0 else b""

            if tag == TAG_START_STREAMING:
                await self._handle_start_streaming(client, payload)
            elif tag == TAG_STOP_STREAMING:
                self._handle_stop_streaming(client, payload)
            elif tag == TAG_TERMINAL_INPUT:
                await self._handle_terminal_input(payload)
            elif tag == TAG_TERMINAL_RESIZE:
                await self._handle_terminal_resize(payload)
            elif tag == TAG_START_TERMINAL:
                await self._handle_start_terminal(payload)
            elif tag == TAG_TERMINAL_SCROLL:
                await self._handle_terminal_scroll(client, payload)
            elif tag == TAG_TERMINAL_CLEAR:
                await self._handle_terminal_clear(client, payload)

    async def _handle_start_streaming(self, client: ClientConnection, payload: bytes) -> None:
        try:
            data = json.loads(payload)
            terminal_id = data["id"]
        except (json.JSONDecodeError, KeyError):
            logger.warning("Malformed start streaming payload, ignoring")
            return
        client.subscribe(terminal_id)

        terminal = self._terminals.get(terminal_id)
        if terminal is not None and terminal.screen is not None:
            sb = terminal.screen.scrollback()
            id_prefix = json.dumps({"id": terminal_id, "sb": sb}).encode() + b"\x00"
            await client.send(encode_tlv(TAG_TERMINAL_CELLS, id_prefix + terminal.screen.cells()))
            return

        bg = self._bg_terminals.get(terminal_id)
        if bg is not None and bg.screen is not None:
            bg.read_new_output()
            sb = bg.screen.scrollback()
            id_prefix = json.dumps({"id": terminal_id, "sb": sb}).encode() + b"\x00"
            await client.send(encode_tlv(TAG_TERMINAL_CELLS, id_prefix + bg.screen.cells()))

    def _handle_stop_streaming(self, client: ClientConnection, payload: bytes) -> None:
        try:
            data = json.loads(payload)
            terminal_id = data["id"]
        except (json.JSONDecodeError, KeyError):
            logger.warning("Malformed stop streaming payload, ignoring")
            return
        client.unsubscribe(terminal_id)

    async def _handle_terminal_input(self, payload: bytes) -> None:
        if self._on_input is None:
            return
        try:
            null_idx = payload.index(0)
            meta = json.loads(payload[:null_idx])
            input_data = payload[null_idx + 1 :].decode("utf-8")
        except (ValueError, json.JSONDecodeError, KeyError, UnicodeDecodeError):
            logger.warning("Malformed terminal input payload, ignoring")
            return
        await self._on_input(meta["id"], input_data)

    async def _handle_terminal_resize(self, payload: bytes) -> None:
        try:
            meta = json.loads(payload)
        except (json.JSONDecodeError, KeyError):
            logger.warning("Malformed terminal resize payload, ignoring")
            return
        terminal_id = meta["id"]
        cols = meta["cols"]
        rows = meta["rows"]
        terminal = self._terminals.get(terminal_id)
        if terminal is not None:
            if terminal.screen is not None:
                terminal.screen.resize(rows, cols)
            terminal.rows = rows
            terminal.cols = cols
        if self._on_resize is not None:
            await self._on_resize(terminal_id, cols, rows)

        handle = self._flush_handles.pop(terminal_id, None)
        if handle is not None:
            handle.cancel()
        self._pending_output.pop(terminal_id, None)

        if terminal is not None and terminal.screen is not None:
            sb = terminal.screen.scrollback()
            id_prefix = json.dumps({"id": terminal_id, "sb": sb}).encode() + b"\x00"
            await self._broadcast_to_subscribers(
                terminal_id, encode_tlv(TAG_TERMINAL_CELLS, id_prefix + terminal.screen.cells())
            )

    async def _handle_terminal_scroll(self, client: ClientConnection, payload: bytes) -> None:
        try:
            meta = json.loads(payload)
            terminal_id = meta["id"]
            offset = meta["offset"]
        except (json.JSONDecodeError, KeyError):
            logger.warning("Malformed terminal scroll payload, ignoring")
            return
        terminal = self._terminals.get(terminal_id)
        if terminal is None or terminal.screen is None:
            return
        terminal.screen.set_scrollback(offset)
        sb = terminal.screen.scrollback()
        id_prefix = json.dumps({"id": terminal_id, "sb": sb}).encode() + b"\x00"
        await client.send(encode_tlv(TAG_TERMINAL_CELLS, id_prefix + terminal.screen.cells()))

    async def _handle_start_terminal(self, payload: bytes) -> None:
        if self._on_start_terminal is None:
            return
        try:
            meta = json.loads(payload)
        except (json.JSONDecodeError, KeyError):
            logger.warning("Malformed start terminal payload, ignoring")
            return
        await self._on_start_terminal(meta["id"], meta["cols"], meta["rows"])

    async def _handle_terminal_clear(self, client: ClientConnection, payload: bytes) -> None:
        try:
            meta = json.loads(payload)
            terminal_id = meta["id"]
        except (json.JSONDecodeError, KeyError):
            logger.warning("Malformed terminal clear payload, ignoring")
            return
        terminal = self._terminals.get(terminal_id)
        if terminal is None:
            return
        handle = self._flush_handles.pop(terminal_id, None)
        if handle is not None:
            handle.cancel()
        self._pending_output.pop(terminal_id, None)
        if terminal.screen is None:
            return
        terminal.screen.process(b"\x1b[H\x1b[2J")
        sb = terminal.screen.scrollback()
        id_prefix = json.dumps({"id": terminal_id, "sb": sb}).encode() + b"\x00"
        await self._broadcast_to_subscribers(
            terminal_id, encode_tlv(TAG_TERMINAL_CELLS, id_prefix + terminal.screen.cells())
        )



import asyncio
import json
import logging
import socket
import uuid
from dataclasses import dataclass

import click
import websockets.asyncio.client
from aiortc import (
    RTCConfiguration,
    RTCIceCandidate,
    RTCIceServer,
    RTCPeerConnection,
    RTCSessionDescription,
)
from aiortc.sdp import candidate_from_sdp

from exponent.commands.common import redirect_to_login, run_until_complete
from exponent.commands.settings import use_settings
from exponent.commands.types import exponent_cli_group
from exponent.core.config import Settings

logger = logging.getLogger(__name__)

BUFFER_SIZE = 16384
DEFAULT_STUN_SERVERS = [
    RTCIceServer(urls=["stun:stun.cloudflare.com:3478"]),
]


def _parse_ice_candidate(candidate_data: dict[str, object]) -> RTCIceCandidate | None:
    candidate_str = str(candidate_data.get("candidate", ""))
    if not candidate_str:
        return None
    candidate = candidate_from_sdp(candidate_str)
    candidate.sdpMid = candidate_data.get("sdpMid")  # type: ignore[assignment]
    candidate.sdpMLineIndex = candidate_data.get("sdpMLineIndex")  # type: ignore[assignment]
    return candidate


@dataclass
class PortSpec:
    local_port: int
    remote_port: int | None
    remote_unix_socket: str | None

    def dc_label(self) -> str:
        if self.remote_unix_socket:
            return f"unix:{self.remote_unix_socket}"
        return f"tcp:{self.remote_port}"


def _validate_port(port: int) -> None:
    if port < 1 or port > 65535:
        raise ValueError(f"port {port} out of range (1-65535)")


def parse_port_spec(spec: str) -> PortSpec:
    if ":" in spec:
        local_str, remote = spec.split(":", 1)
        local_port = int(local_str)
        _validate_port(local_port)
        if remote.startswith("/") or remote.startswith("."):
            return PortSpec(local_port=local_port, remote_port=None, remote_unix_socket=remote)
        else:
            remote_port = int(remote)
            _validate_port(remote_port)
            return PortSpec(local_port=local_port, remote_port=remote_port, remote_unix_socket=None)
    else:
        port = int(spec)
        _validate_port(port)
        return PortSpec(local_port=port, remote_port=port, remote_unix_socket=None)


def _bind_socket(family: socket.AddressFamily, addr: str, port: int) -> socket.socket:
    sock = socket.socket(family, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    if family == socket.AF_INET6:
        sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 1)
    sock.setblocking(False)
    sock.bind((addr, port))
    sock.listen(128)
    return sock


def bind_listeners(specs: list[PortSpec]) -> list[tuple[PortSpec, socket.socket]]:
    listeners: list[tuple[PortSpec, socket.socket]] = []
    for spec in specs:
        listeners.append((spec, _bind_socket(socket.AF_INET, "127.0.0.1", spec.local_port)))
        try:
            listeners.append((spec, _bind_socket(socket.AF_INET6, "::1", spec.local_port)))
        except OSError:
            logger.debug("IPv6 not available for port %d, skipping", spec.local_port)
    return listeners


async def _signaling_handshake(
    ws_url: str,
    api_key: str,
    pc: RTCPeerConnection,
    first_spec: PortSpec,
) -> None:
    ws = await websockets.asyncio.client.connect(ws_url, open_timeout=10)
    try:
        await ws.send(json.dumps({"type": "auth", "data": {"api_key": api_key}}))
        resp = json.loads(await ws.recv())
        if resp.get("type") != "auth_success":
            raise click.ClickException(f"Signaling auth failed: {resp.get('reason', 'unknown')}")

        offer = await pc.createOffer()
        await pc.setLocalDescription(offer)
        while pc.iceGatheringState != "complete":
            await asyncio.sleep(0.05)

        local_desc = pc.localDescription
        assert local_desc is not None

        session_id = str(uuid.uuid4())
        offer_msg: dict[str, object] = {
            "type": "offer",
            "session_id": session_id,
            "payload": json.dumps({"type": "offer", "sdp": local_desc.sdp}),
        }
        if first_spec.remote_port is not None:
            offer_msg["target_port"] = first_spec.remote_port
        if first_spec.remote_unix_socket is not None:
            offer_msg["target_unix_socket"] = first_spec.remote_unix_socket
        await ws.send(json.dumps(offer_msg))

        ice_connected = asyncio.Event()
        got_answer = False

        def _check_ice_state() -> None:
            state = pc.iceConnectionState
            logger.debug("ICE connection state: %s", state)
            if state in ("connected", "completed"):
                ice_connected.set()
            elif state == "failed":
                ice_connected.set()

        pc.on("iceconnectionstatechange", _check_ice_state)
        _check_ice_state()

        async def _read_signaling() -> None:
            nonlocal got_answer
            async for raw in ws:
                msg = json.loads(raw)
                msg_type = msg.get("type")

                if msg_type == "session_ok":
                    continue

                if msg_type == "signal":
                    kind = msg.get("kind")
                    payload = str(msg.get("payload", ""))

                    if kind == "ANSWER":
                        answer_data = json.loads(payload)
                        await pc.setRemoteDescription(
                            RTCSessionDescription(sdp=str(answer_data.get("sdp", "")), type="answer")
                        )
                        got_answer = True
                        click.echo("  Received answer, waiting for ICE...")

                    elif kind == "ICE_CANDIDATE":
                        candidate = _parse_ice_candidate(json.loads(payload))
                        if candidate:
                            await pc.addIceCandidate(candidate)

                elif msg_type == "error":
                    raise click.ClickException(f"Signaling error: {msg.get('error', 'unknown')}")

                elif msg_type == "ping":
                    await ws.send(json.dumps({"type": "pong"}))

            if not got_answer:
                raise click.ClickException("Signaling WebSocket closed before answer received")

        read_task = asyncio.create_task(_read_signaling())
        done, _pending = await asyncio.wait(
            [read_task, asyncio.create_task(ice_connected.wait())],
            timeout=30.0,
            return_when=asyncio.FIRST_COMPLETED,
        )

        read_task.cancel()
        try:
            await read_task
        except asyncio.CancelledError:
            pass  # Expected: we just cancelled read_task above

        if not done:
            if got_answer:
                raise click.ClickException(f"Timed out waiting for ICE connection (state: {pc.iceConnectionState})")
            raise click.ClickException("Timed out waiting for signaling answer from sandbox")

        for task in done:
            exc = task.exception()
            if exc is not None:
                raise exc

        if pc.iceConnectionState == "failed":
            raise click.ClickException("ICE connection failed")
    finally:
        await ws.close()


async def _fetch_ice_servers(ws_url: str, api_key: str) -> list[RTCIceServer]:
    ws = await websockets.asyncio.client.connect(ws_url, open_timeout=10)
    try:
        await ws.send(json.dumps({"type": "auth", "data": {"api_key": api_key}}))
        resp = json.loads(await ws.recv())
        if resp.get("type") != "auth_success":
            raise click.ClickException(f"Signaling auth failed: {resp.get('reason', 'unknown')}")

        ice_servers: list[RTCIceServer] = []
        for cfg in resp.get("ice_servers") or []:
            ice_servers.append(
                RTCIceServer(
                    urls=cfg["urls"],
                    username=cfg.get("username", ""),
                    credential=cfg.get("credential", ""),
                )
            )
        if not ice_servers:
            return list(DEFAULT_STUN_SERVERS)
        return ice_servers
    finally:
        await ws.close()


async def _bridge_connection(
    reader: asyncio.StreamReader,
    writer: asyncio.StreamWriter,
    pc: RTCPeerConnection,
    spec: PortSpec,
) -> None:
    peer_addr = writer.get_extra_info("peername")
    target_label = spec.remote_unix_socket or str(spec.remote_port)
    click.echo(f"New connection on :{spec.local_port} from {peer_addr} -> {target_label}")

    dc = pc.createDataChannel(spec.dc_label(), ordered=True, protocol="bridge-v1")
    done = asyncio.Event()
    dc_open = asyncio.Event()
    bridge_ready = asyncio.Event()
    remote_error: str | None = None

    @dc.on("open")
    def on_open() -> None:
        dc_open.set()

    @dc.on("close")
    def on_close() -> None:
        done.set()

    @dc.on("message")
    def on_message(data: bytes | str) -> None:
        nonlocal remote_error
        if isinstance(data, str):
            data = data.encode()
        if not bridge_ready.is_set():
            if len(data) == 0:
                return
            if data[0] == 0x00:
                bridge_ready.set()
            elif data[0] == 0x01:
                remote_error = data[1:].decode(errors="replace")
                done.set()
            return

        try:
            writer.write(data)
        except Exception:
            done.set()

    try:
        await asyncio.wait_for(dc_open.wait(), timeout=10.0)
    except TimeoutError:
        click.echo(f"Timed out waiting for data channel open on :{spec.local_port}", err=True)
        dc.close()
        writer.close()
        return

    async def tcp_to_dc() -> None:
        try:
            while not done.is_set():
                data = await reader.read(BUFFER_SIZE)
                if not data:
                    break
                dc.send(data)
        except Exception:
            logger.debug("tcp_to_dc error on :%d", spec.local_port, exc_info=True)
        finally:
            done.set()

    tcp_task = asyncio.create_task(tcp_to_dc())
    await done.wait()
    tcp_task.cancel()
    try:
        await tcp_task
    except asyncio.CancelledError:
        pass  # Expected: we just cancelled the task above

    dc.close()
    writer.close()
    try:
        await writer.wait_closed()
    except OSError:
        logger.debug("Error closing writer on :%d", spec.local_port, exc_info=True)

    if remote_error:
        click.echo(f"Connection refused on :{spec.local_port} -> {target_label}: {remote_error}", err=True)
    else:
        click.echo(f"Connection closed on :{spec.local_port} from {peer_addr}")


async def port_forward_async(
    chat_uuid: str,
    specs: list[PortSpec],
    ws_url: str,
    api_key: str,
) -> None:
    listeners = bind_listeners(specs)
    for spec in specs:
        if spec.remote_unix_socket:
            click.echo(f"Forwarding 127.0.0.1:{spec.local_port} -> {spec.remote_unix_socket}")
        else:
            click.echo(f"Forwarding 127.0.0.1:{spec.local_port} -> :{spec.remote_port}")

    click.echo(f"\nConnecting to sandbox {chat_uuid}...")
    ice_servers = await _fetch_ice_servers(ws_url, api_key)

    config = RTCConfiguration(iceServers=ice_servers)
    pc = RTCPeerConnection(configuration=config)
    pc.createDataChannel("_sctp", negotiated=True, id=0)

    stop = asyncio.Event()

    def on_connection_state_change() -> None:
        state = pc.connectionState
        if state in ("failed", "closed"):
            click.echo(f"WebRTC connection {state}, shutting down.", err=True)
            stop.set()

    pc.on("connectionstatechange", on_connection_state_change)

    click.echo("  Establishing WebRTC session...")
    await _signaling_handshake(ws_url, api_key, pc, specs[0])
    click.echo("Connected. Press Ctrl+C to stop.\n")

    servers: list[asyncio.AbstractServer] = []
    for spec, sock in listeners:

        async def on_connect(
            reader: asyncio.StreamReader,
            writer: asyncio.StreamWriter,
            _pc: RTCPeerConnection = pc,
            _spec: PortSpec = spec,
        ) -> None:
            try:
                await _bridge_connection(reader, writer, _pc, _spec)
            except Exception:
                logger.exception("error handling connection on :%d", _spec.local_port)

        server = await asyncio.start_server(on_connect, sock=sock)
        servers.append(server)

    try:
        await stop.wait()
    except asyncio.CancelledError:
        logger.debug("port-forward shutting down")
    finally:
        for server in servers:
            server.close()
        for server in servers:
            await server.wait_closed()
        await pc.close()


@exponent_cli_group()
def port_forward_cli() -> None:
    pass


@port_forward_cli.command("port-forward")
@click.argument("chat_uuid")
@click.argument("ports", nargs=-1, required=True)
@use_settings
def port_forward(settings: Settings, chat_uuid: str, ports: tuple[str, ...]) -> None:
    """Forward local ports to a sandbox via WebRTC.

    CHAT_UUID is the chat/sandbox identifier.

    PORTS are port specs in the form:
      3002          - forward local 3002 to remote 3002
      8000:8080     - forward local 8000 to remote 8080
      2222:/tmp/s   - forward local 2222 to a remote unix socket
    """
    if not settings.api_key:
        redirect_to_login(settings)
        return

    specs: list[PortSpec] = []
    for p in ports:
        try:
            specs.append(parse_port_spec(p))
        except ValueError:
            raise click.BadParameter(f"Invalid port spec: {p}")

    ws_base = settings.get_base_ws_url()
    ws_url = f"{ws_base}/api/ws/webrtc/{chat_uuid}"

    run_until_complete(port_forward_async(chat_uuid, specs, ws_url, settings.api_key))



# Generated by ariadne-codegen
# Source: exponent/core/graphql/operations

from typing import Literal

from pydantic import Field

from .base_model import BaseModel


class RefreshApiKey(BaseModel):
    refresh_api_key: "RefreshApiKeyRefreshApiKey" = Field(alias="refreshApiKey")


class RefreshApiKeyRefreshApiKey(BaseModel):
    typename__: Literal["User"] = Field(alias="__typename")
    user_api_key: str = Field(alias="userApiKey")


RefreshApiKey.model_rebuild()
